@@ -1,5 +1,5 @@
 # -*- coding: UTF-8 -*-
-## Copyright 2010-2012 Luc Saffre
+## Copyright 2010-2013 Luc Saffre
 ## This file is part of the Lino project.
 ## Lino is free software; you can redistribute it and/or modify
 ## it under the terms of the GNU General Public License as published by
@@ -24,7 +24,7 @@ import codecs
 import time
 import datetime
 #~ import signal
-import atexit
+#~ import atexit
 
 #~ import logging
 #~ logger = logging.getLogger(__name__)
@@ -37,7 +37,6 @@ from django.core.exceptions import ValidationError
 
 from django.conf import settings
 
-
 from django.db.utils import DatabaseError
 # OperationalError
 from django.utils import simplejson
@@ -54,26 +53,43 @@ from lino.utils import join_words, unicode_string
 from lino.utils import dblogger
 from lino.utils import mti
 from lino.core.modeltools import obj2str
-from lino.core import changes
+#~ from lino.core import changes
+#~ from lino.modlib import models as changes
 from lino.core.modeltools import is_valid_email
 
 from lino.utils.daemoncommand import DaemonCommand
 
 from lino.utils.ssin import is_valid_ssin
 
-
 #~ from lino_welfare.modlib.pcsw import models as pcsw
 #~ from lino_welfare.modlib.pcsw.management.commands.initdb_tim import ADR_id
 
+settings.LINO.startup()
+
 pcsw = dd.resolve_app('pcsw')
 users = dd.resolve_app('users')
+#~ contacts = dd.resolve_app('contacts')
+#~ households = dd.resolve_app('households')
+#~ countries = dd.resolve_app('countries')
+
+#~ Country = countries.Country
+#~ City = countries.City
 Country = dd.resolve_model('countries.Country')
 City = dd.resolve_model('countries.City')
-Person = dd.resolve_model(settings.LINO.person_model)
+#~ Person = contacts.Person
+#~ Person = pcsw.Person
+#~ Company = contacts.Company
+#~ Company = pcsw.Company
+Person = dd.resolve_model('contacts.Person')
 Client = dd.resolve_model('pcsw.Client')
-Company = dd.resolve_model(settings.LINO.company_model)
+Partner = dd.resolve_model('contacts.Partner')
+#~ Client = pcsw.Client
+Company = dd.resolve_model('contacts.Company')
 Household = dd.resolve_model('households.Household')
 households_Type = dd.resolve_model("households.Type")
+#~ Household = pcsw.Household 
+#~ Household = households.Household 
+#~ households_Type = households.Type
 
 CCTYPE_HEALTH_INSURANCE = 1
 CCTYPE_PHARMACY = 2
@@ -146,14 +162,16 @@ def checkcc(person,pk,nType):
             company_id=pk,
             type=pcsw.ClientContactType.objects.get(id=nType))
         cc.save()
-        changes.log_create(REQUEST,cc)
+        dd.pre_ui_create.send(sender=cc,request=REQUEST)
+        #~ changes.log_create(REQUEST,cc)
     elif qs.count() == 1:
         cc = qs[0]
         if cc.company_id != pk:
-            watcher = changes.Watcher(cc)
+            watcher = dd.ChangeWatcher(cc)
             cc.company_id = pk
             cc.save()
-            watcher.log_diff(REQUEST)
+            watcher.send_update(REQUEST)
+            #~ watcher.log_diff(REQUEST)
     else:
         dblogger.warning(u"%s : more than 1 ClientContact (type=%r)",obj2str(person),nType)
 
@@ -282,6 +300,8 @@ def PAR_model(data):
     attribs = data.get('ATTRIB',False)
     if attribs and 'N' in attribs: # newcomer
         return Client
+    if data.get('IDUSR',False): # Sozi
+        return Client
     if data.get('NOTVA',False):
         return Company
     if data.get('ALLO','') in (u"Eheleute",):
@@ -311,12 +331,7 @@ def json2py(dct):
 CONTACT_FIELDS = '''id name street street_no street_box addr2 
 country city zip_code region language email url phone gsm remarks'''.split()
 
-#~ class PseudoRequest:
-    #~ user = "watch_tim"
-#~ REQUEST = PseudoRequest()
-
-REQUEST = changes.PseudoRequest("watch_tim")
-#~ 20120921 REQUEST = dblogger.PseudoRequest("watch_tim")
+REQUEST = dd.PseudoRequest("watch_tim")
 
 class Controller:
     "Deserves more documentation."
@@ -335,7 +350,7 @@ class Controller:
         
     def validate_and_save(self,obj):
         "Deserves more documentation."
-        dblogger.info("20121022 validate_and_save %s",obj2str(obj,True))
+        #~ dblogger.info("20121022 validate_and_save %s",obj2str(obj,True))
         obj.full_clean()
         #~ 20120921 dblogger.log_changes(REQUEST,obj)
         obj.save()
@@ -368,7 +383,8 @@ class Controller:
             return
             
         #~ 20120921 dblogger.log_deleted(REQUEST,obj)
-        changes.log_delete(REQUEST,obj)
+        dd.pre_ui_delete.send(sender=obj,request=REQUEST)
+        #~ changes.log_delete(REQUEST,obj)
         obj.delete()
         dblogger.debug("%s:%s (%s) : DELETE ok",kw['alias'],kw['id'],obj2str(obj))
         
@@ -392,14 +408,16 @@ class Controller:
             self.applydata(obj,kw['data'])
             dblogger.debug("%s:%s (%s) : POST %s",kw['alias'],kw['id'],obj2str(obj),kw['data'])
             self.validate_and_save(obj)
-            changes.log_create(REQUEST,obj)
+            dd.pre_ui_create.send(sender=obj,request=REQUEST)
+            #~ changes.log_create(REQUEST,obj)
         else:
-            watcher = changes.Watcher(obj)
+            watcher = dd.ChangeWatcher(obj)
             dblogger.info("%s:%s : POST becomes PUT",kw['alias'],kw['id'])
             self.applydata(obj,kw['data'])
             dblogger.debug("%s:%s (%s) : POST %s",kw['alias'],kw['id'],obj2str(obj),kw['data'])
             self.validate_and_save(obj)
-            watcher.log_diff(REQUEST)
+            watcher.send_update(REQUEST)
+            #~ watcher.log_diff(REQUEST)
             #~ obj.save()
         
     def PUT(self,**kw):
@@ -413,13 +431,14 @@ class Controller:
             else:
                 dblogger.warning("%s:%s : PUT ignored (row does not exist)",kw['alias'],kw['id'])
                 return 
-        watcher = changes.Watcher(obj)
+        watcher = dd.ChangeWatcher(obj)
         if self.PUT_special(watcher,**kw):
             return 
         self.applydata(obj,kw['data'])
         dblogger.debug("%s:%s (%s) : PUT %s",kw['alias'],kw['id'],obj2str(obj),kw['data'])
         self.validate_and_save(obj)
-        watcher.log_diff(REQUEST)
+        watcher.send_update(REQUEST)
+        #~ watcher.log_diff(REQUEST)
         #~ obj.save()
         #~ dblogger.debug("%s:%s : PUT %s",kw['alias'],kw['id'],kw['data'])
         
@@ -498,67 +517,94 @@ class PAR(Controller):
                 elif data['IDPRT'] == 'I':
                     obj.client_state = pcsw.ClientStates.former
                 #~ 
-                elif obj.national_id and is_valid_ssin(obj.national_id):
-                    obj.client_state = pcsw.ClientStates.coached
                 else:
-                    obj.client_state = pcsw.ClientStates.invalid
+                    obj.client_state = pcsw.ClientStates.coached
+                #~ elif obj.national_id and is_valid_ssin(obj.national_id):
+                    #~ obj.client_state = pcsw.ClientStates.coached
+                #~ else:
+                    #~ obj.client_state = pcsw.ClientStates.invalid
                 #~ if data.has_key('NB1'):
                     #~ obj.gesdos_id = data['NB1']
                 #~ if not obj.national_id:
                     #~ obj.national_id = str()
                 if data.has_key('IDUSR'):
                     username = settings.TIM2LINO_USERNAME(data['IDUSR'])
-                    if obj.pk is None:
-                        obj.save() # must force a pre save() here to save related coachings
                     if username:
+                        #~ print 20130222, username
                         u = users.User.objects.get(username=username)
-                        #~ try:
-                            #~ u = users.User.objects.get(username=username)
-                        #~ except users.User.DoesNotExist,e:
-                            #~ dblogger.warning(
-                              #~ u"PAR->IdUsr %r (converted to %r) doesn't exist! while saving %s",
-                              #~ data['IDUSR'],username,obj2str(obj))
-                        #~ obj.coach1 = u
                         """
                         typical cases:
-                        - imported client has got assign_coach in Lino, then removed PARATTR_N in TIM
+                        - imported client has been assigned a coach in Lino, 
+                          then filled IDUSR and removed PARATTR_N in TIM
                         """
-                        try:
-                            coaching = pcsw.Coaching.objects.get(client=obj,primary=True)
-                            #~ if coaching.user != u or coaching.start_date != obj.created or coaching.end_date is not None:
-                            if coaching.user != u or coaching.end_date is not None or coaching.start_date is None:
-                                watcher = changes.Watcher(coaching)
-                                coaching.user = u
+                    else:
+                        u = None
+                        
+                    if obj.pk is None:
+                        obj.save() # must pre-save the client here to save related coachings
+                        
+                    try:
+                        coaching = pcsw.Coaching.objects.get(client=obj,primary=True)
+                        #~ if coaching.user != u or coaching.start_date != obj.created or coaching.end_date is not None:
+                        if coaching.user == u:
+                            if coaching.end_date is not None \
+                                    or coaching.start_date is None\
+                                    or coaching.type != u.coaching_type:
+                                watcher = dd.ChangeWatcher(coaching)
+                                coaching.type = u.coaching_type
                                 if coaching.start_date is None:
                                     coaching.start_date = obj.created
                                 coaching.end_date = None
                                 coaching.save()
-                                watcher.log_diff(REQUEST)
-                        except pcsw.Coaching.DoesNotExist,e:
-                            try:
-                                coaching = pcsw.Coaching.objects.get(client=obj,user=u,end_date__isnull=True)
-                                watcher = changes.Watcher(coaching)
-                                coaching.primary = True
-                                coaching.save()
-                                watcher.log_diff(REQUEST)
-                            except pcsw.Coaching.DoesNotExist,e:
-                                coaching = pcsw.Coaching(client=obj,primary=True,user=u,start_date=obj.created)
-                                coaching.save()
-                                changes.log_create(REQUEST,coaching)
-                            except Exception,e:
-                                raise Exception("More than one primary coaching for %r by %r" % (obj,u))
-                        except Exception,e:
-                            raise Exception("More than one primary coaching for %r : %s" % (obj,e))
-
-                    else:
-                        #~ obj.coach1 = None
-                        try:
-                            coaching = pcsw.Coaching.objects.get(client=obj,primary=True)
-                            changes.log_delete(REQUEST,coaching)
-                            coaching.delete()
+                                watcher.send_update(REQUEST)
+                                #~ watcher.log_diff(REQUEST)
+                        else:
+                            """
+                            If the coach has changed, maintain the old coaching in history.
+                            """
+                            watcher = dd.ChangeWatcher(coaching)
+                            coaching.primary = False
+                            coaching.end_date = datetime.date.today()
+                            coaching.save()
+                            watcher.send_update(REQUEST)
                             
+                            if u is not None:
+                                coaching = pcsw.Coaching(
+                                    client=obj,primary=True,
+                                    user=u,
+                                    type=u.coaching_type,
+                                    start_date=datetime.date.today())
+                                coaching.save()
+                                dd.pre_ui_create.send(sender=coaching,request=REQUEST)
+
+                    except pcsw.Coaching.DoesNotExist,e:
+                        try:
+                            coaching = pcsw.Coaching.objects.get(client=obj,user=u,end_date__isnull=True)
+                            watcher = dd.ChangeWatcher(coaching)
+                            coaching.primary = True
+                            coaching.save()
+                            watcher.send_update(REQUEST)
+                            #~ watcher.log_diff(REQUEST)
                         except pcsw.Coaching.DoesNotExist,e:
-                            pass
+                            if u is not None:
+                                coaching = pcsw.Coaching(client=obj,primary=True,user=u,
+                                  type=u.coaching_type,
+                                  start_date=obj.created)
+                                coaching.save()
+                                dd.pre_ui_create.send(sender=coaching,request=REQUEST)
+                                #~ changes.log_create(REQUEST,coaching)
+                        except Exception,e:
+                            raise Exception("More than one primary coaching for %r by %r" % (obj,u))
+                    except Exception,e:
+                        raise Exception("More than one primary coaching for %r : %s" % (obj,e))
+
+                    #~ else:
+                        #~ try:
+                            #~ coaching = pcsw.Coaching.objects.get(client=obj,primary=True)
+                            #~ dd.pre_ui_delete.send(sender=coaching,request=REQUEST)
+                            #~ coaching.delete()
+                        #~ except pcsw.Coaching.DoesNotExist,e:
+                            #~ pass
                         
         elif obj.__class__ is Company:
             mapper.update(prefix='ALLO')
@@ -600,6 +646,10 @@ class PAR(Controller):
             return Household.objects.get(pk=id)
         except Household.DoesNotExist:
             pass
+        try:
+            return Partner.objects.get(pk=id)
+        except Partner.DoesNotExist:
+            pass
         
     def PUT_special(self,watcher,**kw):
         obj = watcher.watched
@@ -619,26 +669,43 @@ class PAR(Controller):
             #~ if v is not None:
                 #~ kw[n] = v
         old_class = obj.__class__
+        #~ print 20130222, old_class, new_class
         if issubclass(old_class,new_class):
             #~ it was a Client and becomes a Person
-            changes.log_remove_child(REQUEST,obj,old_class)
+            dd.pre_remove_child.send(sender=obj,request=REQUEST,child=old_class)
+            #~ changes.log_remove_child(REQUEST,obj,old_class)
             mti.delete_child(obj,old_class)
             #~ changes.log_delete(REQUEST,obj)
             newobj = new_class.objects.get(pk=obj.id)
         elif issubclass(new_class,old_class):
             #~ it was only a Person and becomes a Client
-            changes.log_add_child(REQUEST,obj,new_class)
+            dd.pre_add_child.send(sender=obj,request=REQUEST,child=new_class)
             newobj = mti.insert_child(obj,new_class)
+        elif new_class is Client and old_class is Company:
+            """
+            It's a Company and must become a Client (and remain a Company).
+            We must first create a Person.
+            """
+            partner = obj.partner_ptr
+            # create and save the Person:
+            dd.pre_add_child.send(sender=partner,request=REQUEST,child=Person)
+            person = mti.insert_child(partner,Person)
+            self.applydata(person,data)
+            self.validate_and_save(person)
+            # create the Client
+            dd.pre_add_child.send(sender=person,request=REQUEST,child=new_class)
+            newobj = mti.insert_child(person,new_class)
         else:
+            raise Exception("Cannot handle conversion from %s to %s" % (old_class,new_class))
             obj = obj.partner_ptr
-            changes.log_remove_child(REQUEST,obj,old_class)
+            dd.pre_remove_child.send(sender=obj,request=REQUEST,child=old_class)
             mti.delete_child(obj,old_class)
-            changes.log_add_child(REQUEST,obj,new_class)
+            dd.pre_add_child.send(sender=obj,request=REQUEST,child=new_class)
             newobj = mti.insert_child(obj,new_class)
             
         self.applydata(newobj,data)
         self.validate_and_save(newobj)
-        return newobj
+        #~ return newobj
         
     def create_object(self,kw):
         return PAR_model(kw['data'])(id=kw['id'])
@@ -670,7 +737,7 @@ class PXS(Controller):
     """
     
     def create_object(self,kw):
-        raise Exception("Tried to create a Person from PXS")
+        raise Exception("Cannot create Client %s from PXS" % kw['id'])
         
     def PUT_special(self,watcher,**kw):
         pass
@@ -788,8 +855,7 @@ controllers = dict(
   ADR=ADR(),
   )
 
-def process_line(i,ln):
-    dblogger.debug("process_line(%r,%r)",i,ln)
+def process_line(ln):
     d = simplejson.loads(ln,object_hook=json2py)
     kw = {}
     for k,v in d.items():
@@ -827,8 +893,9 @@ def watch(data_dir):
     i = 0
     for ln in fd_watching.readlines():
         i += 1
+        dblogger.debug("process_line(%r,%r)",i,ln)
         try:
-            process_line(i,ln)
+            process_line(ln)
         except Exception,e:
             #~ raise
             fd_failed.write("// %s %r\n%s\n\n" % (
@@ -853,18 +920,18 @@ def main(*args,**options):
     if len(args) != 1:
         raise CommandError('Please specify the path to your TIM changelog directory')
     data_dir = args[0]
-    msg = "Started watch_tim %s on %s ..."
+    #~ msg = "Started watch_tim %s on %s ..."
     #~ logger.info(msg,data_dir)
-    dblogger.info(msg,lino.__version__,data_dir)
+    #~ dblogger.info(msg,lino.__version__,data_dir)
     
     settings.LINO.startup() 
         
-    def goodbye():
-        msg = "Stopped watch_tim %s on %s ..."
+    #~ def goodbye():
+        #~ msg = "Stopped watch_tim %s on %s ..."
         #~ logger.info(msg,data_dir)
-        dblogger.info(msg,lino.__version__,data_dir)
-    #~ signal.signal(signal.SIGTERM,on_SIGTERM)
-    atexit.register(goodbye)
+        #~ dblogger.info(msg,lino.__version__,data_dir)
+    #~ # signal.signal(signal.SIGTERM,on_SIGTERM)
+    #~ atexit.register(goodbye)
     
     #~ last_warning = None
     while True:
